@@ -12,6 +12,12 @@ import warnings
 import hashlib
 from contextlib import contextmanager
 from .data.dian import codelist
+from . import form
+from collections import defaultdict
+
+AMBIENTE_PRUEBAS = codelist.TipoAmbiente.by_name('Pruebas')['code']
+AMBIENTE_PRODUCCION = codelist.TipoAmbiente.by_name('Producción')['code']
+
 
 SCHEME_AGENCY_ATTRS = {
     'schemeAgencyName': 'CO, DIAN (Dirección de Impuestos y Aduanas Nacionales)',
@@ -25,6 +31,7 @@ POLICY_NAME = u'Política de firma para facturas electrónicas de la República
 
 
 NAMESPACES = {
+    'facho': 'http://git.disroot.org/Etrivial/facho',
     'fe': 'http://www.dian.gov.co/contratos/facturaelectronica/v1',
     'cac': 'urn:oasis:names:specification:ubl:schema:xsd:CommonAggregateComponents-2',
     'cbc': 'urn:oasis:names:specification:ubl:schema:xsd:CommonBasicComponents-2',
@@ -33,7 +40,7 @@ NAMESPACES = {
     'clmIANAMIMEMediaType': 'urn:un:unece:uncefact:codelist:specification:IANAMIMEMediaType:2003',
     'ext': 'urn:oasis:names:specification:ubl:schema:xsd:CommonExtensionComponents-2',
     'qdt': 'urn:oasis:names:specification:ubl:schema:xsd:QualifiedDatatypes-2',
-    'sts': 'http://www.dian.gov.co/contratos/facturaelectronica/v1/Structures',
+    'sts': 'dian:gov:co:facturaelectronica:Structures-2-1',
     'udt': 'urn:un:unece:uncefact:data:specification:UnqualifiedDataTypesSchemaModule:2',
     'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
     'xades': 'http://uri.etsi.org/01903/v1.3.2#',
@@ -41,6 +48,10 @@ NAMESPACES = {
     'sig': 'http://www.w3.org/2000/09/xmldsig#',
 }
 
+def fe_from_string(document: str) -> FachoXML:
+    xml = LXMLBuilder.from_string(document)
+    return FachoXML(xml, nsmap=NAMESPACES)
+
 from contextlib import contextmanager
 @contextmanager
 def mock_xades_policy():
@@ -68,35 +79,45 @@ class FeXML(FachoXML):
         self._cn = root.rstrip('/')
         #self.find_or_create_element(self._cn)
 
-    # MACHETE se elimina xml namespace fe
-    def tostringMACHETE(self, **kw):
+    def tostring(self, **kw):
         return super().tostring(**kw)\
             .replace("fe:", "")\
             .replace("xmlns:fe", "xmlns")
 
 
-class DianXMLExtensionCUFE(FachoXMLExtension):
-    AMBIENTE_PRUEBAS = codelist.TipoAmbiente.by_name('Pruebas')['code']
-    AMBIENTE_PRODUCCION = codelist.TipoAmbiente.by_name('Producción')['code']
 
-    def __init__(self, invoice, tipo_ambiente = AMBIENTE_PRUEBAS, clave_tecnica = ''):
+class DianXMLExtensionCUDFE(FachoXMLExtension):
+
+    def __init__(self, invoice, tipo_ambiente = AMBIENTE_PRUEBAS):
         self.tipo_ambiente = tipo_ambiente
-        self.clave_tecnica = clave_tecnica
         self.invoice = invoice
 
-    def _tipo_ambiente(self):
+    def _tipo_ambiente_int(self):
         return int(self.tipo_ambiente)
 
+    def formatVars(self, invoice):
+        raise NotImplementedError()
+
+    def schemeName(self):
+        raise NotImplementedError()
+
+    def _get_qrcode(self, cufe):
+        url_for = {
+            AMBIENTE_PRUEBAS: 'https://catalogo-vpfe-hab.dian.gov.co/document/searchqr?documentkey=',
+            AMBIENTE_PRODUCCION: 'https://catalogo-vpfe.dian.gov.co/document/searchqr?documentkey='
+        }
+        return url_for[self.tipo_ambiente] + cufe
+
     def build(self, fachoxml):
-        cufe = self._generate_cufe(self.invoice, fachoxml)
-        fachoxml.set_element('/fe:Invoice/cbc:UUID', cufe,
+        cufe = self._generate_cufe()
+        fachoxml.set_element('./cbc:UUID', cufe,
                              schemeID=self.tipo_ambiente,
-                             schemeName='CUFE-SHA384')
-        fachoxml.set_element('/fe:Invoice/cbc:ProfileID', 'DIAN 2.1')
-        fachoxml.set_element('/fe:Invoice/cbc:ProfileExecutionID', self._tipo_ambiente())
+                             schemeName=self.schemeName())
+        fachoxml.set_element('./cbc:ProfileID', 'DIAN 2.1')
+        fachoxml.set_element('./cbc:ProfileExecutionID', self._tipo_ambiente_int())
         #DIAN 1.7.-2020: FAB36
-        fachoxml.set_element('/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:QRCode',
-                'https://catalogo-vpfe.dian.gov.co/document/searchqr?documentkey='+cufe)
+        fachoxml.set_element('./ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:QRCode',
+                self._get_qrcode(cufe))
 
     def issue_time(self, datetime_):
         return datetime_.strftime('%H:%M:%S-05:00')
@@ -104,53 +125,36 @@ class DianXMLExtensionCUFE(FachoXMLExtension):
     def issue_date(self, datetime_):
         return datetime_.strftime('%Y-%m-%d')
 
-    def formatVars(self, invoice):
-        NumFac = invoice.invoice_ident
-        FecFac = self.issue_date(invoice.invoice_issue)
-        HoraFac = self.issue_time(invoice.invoice_issue)
+    def buildVars(self):
+        invoice = self.invoice
+        build_vars = {}
+        build_vars['NumFac'] = invoice.invoice_ident
+        build_vars['FecFac'] = self.issue_date(invoice.invoice_issue)
+        build_vars['HoraFac'] = self.issue_time(invoice.invoice_issue)
         # PAG 601
-        ValorBruto = invoice.invoice_legal_monetary_total.line_extension_amount
-        ValorTotalPagar = invoice.invoice_legal_monetary_total.payable_amount
-        ValorImpuestoPara = {}
-        ValorImpuesto1 = 0.0
-        CodImpuesto1 = 1
-        ValorImpuesto2 = 0.0
-        CodImpuesto2 = 4
-        ValorImpuesto3 = 0.0
-        CodImpuesto3 = 3
+        build_vars['ValorBruto'] = invoice.invoice_legal_monetary_total.line_extension_amount
+        build_vars['ValorTotalPagar'] = invoice.invoice_legal_monetary_total.payable_amount
+        ValorImpuestoPara = defaultdict(lambda: form.Amount(0.0))
+        build_vars['CodImpuesto1'] = '01'
+        build_vars['CodImpuesto2'] = '04'
+        build_vars['CodImpuesto3'] = '03'
         for invoice_line in invoice.invoice_lines:
             for subtotal in invoice_line.tax.subtotals:
-                # TODO cual es la naturaleza de tax_scheme_ident?
-                codigo_impuesto = int(subtotal.tax_scheme_ident)
-                ValorImpuestoPara.setdefault(codigo_impuesto, 0.0)
-                ValorImpuestoPara[codigo_impuesto] += subtotal.tax_amount
+                if subtotal.scheme is not None:
+                    # TODO cual es la naturaleza de tax_scheme_ident?
+                    codigo_impuesto = subtotal.scheme.code
+                    ValorImpuestoPara.setdefault(codigo_impuesto, form.Amount(0.0))
+                    ValorImpuestoPara[codigo_impuesto] += subtotal.tax_amount
 
-        NitOFE = invoice.invoice_supplier.ident
-        NumAdq = invoice.invoice_customer.ident
-        TipoAmb = self._tipo_ambiente()
-        ClTec = str(self.clave_tecnica)
+        build_vars['ValorImpuestoPara'] = ValorImpuestoPara
+        build_vars['NitOFE'] = invoice.invoice_supplier.ident
+        build_vars['NumAdq'] = invoice.invoice_customer.ident
+        build_vars['TipoAmb'] = self._tipo_ambiente_int()
 
-        return [
-            '%s' % NumFac,
-            '%s' % FecFac,
-            '%s' % HoraFac,
-            '%.02f' % ValorBruto,
-            '%02d' % CodImpuesto1,
-            '%.02f' % ValorImpuestoPara.get(CodImpuesto1, 0.0),
-            '%02d' % CodImpuesto2,
-            '%.02f' % ValorImpuestoPara.get(CodImpuesto2, 0.0),
-            '%02d' % CodImpuesto3,
-            '%.02f' % ValorImpuestoPara.get(CodImpuesto3, 0.0),
-            '%.02f' % ValorTotalPagar,
-            '%s' % NitOFE,
-            '%s' % NumAdq,
-            '%s' % ClTec,
-            '%d' % TipoAmb,
-        ]
+        return build_vars
 
-    def _generate_cufe(self, invoice, fachoxml):
-        formatVars = self.formatVars(invoice)
-        cufe = "".join(formatVars)
+    def _generate_cufe(self):
+        cufe = "".join(self.formatVars())
 
         # crear hash...
         h = hashlib.sha384()
@@ -158,6 +162,80 @@ class DianXMLExtensionCUFE(FachoXMLExtension):
         return h.hexdigest()
 
 
+class DianXMLExtensionCUFE(DianXMLExtensionCUDFE):
+    def __init__(self, invoice,  clave_tecnica = '', tipo_ambiente = AMBIENTE_PRUEBAS):
+        self.tipo_ambiente = tipo_ambiente
+        self.clave_tecnica = clave_tecnica
+        self.invoice = invoice
+
+    def schemeName(self):
+        return 'CUFE-SHA384'
+
+    def buildVars(self):
+        build_vars = super().buildVars()
+        build_vars['ClTec'] = str(self.clave_tecnica)
+        return build_vars
+
+    def formatVars(self):
+        build_vars = self.buildVars()
+        CodImpuesto1 = build_vars['CodImpuesto1']
+        CodImpuesto2 = build_vars['CodImpuesto2']
+        CodImpuesto3 = build_vars['CodImpuesto3']
+        return [
+            '%s' % build_vars['NumFac'],
+            '%s' % build_vars['FecFac'],
+            '%s' % build_vars['HoraFac'],
+            form.Amount(build_vars['ValorBruto']).format('%.02f'),
+            CodImpuesto1,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto1, 0.0)).format('%.02f'),
+            CodImpuesto2,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto2, 0.0)).format('%.02f'),
+            CodImpuesto3,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto3, 0.0)).format('%.02f'),
+            form.Amount(build_vars['ValorTotalPagar']).format('%.02f'),
+            '%s' % build_vars['NitOFE'],
+            '%s' % build_vars['NumAdq'],
+            '%s' % build_vars['ClTec'],
+            '%d' % build_vars['TipoAmb'],
+        ]
+
+class DianXMLExtensionCUDE(DianXMLExtensionCUDFE):
+    def __init__(self, invoice, software_pin, tipo_ambiente = AMBIENTE_PRUEBAS):
+        self.tipo_ambiente = tipo_ambiente
+        self.software_pin = software_pin
+        self.invoice = invoice
+
+    def schemeName(self):
+        return 'CUDE-SHA384'
+
+    def buildVars(self):
+        build_vars = super().buildVars()
+        build_vars['Software-PIN'] = str(self.software_pin)
+        return build_vars
+
+    def formatVars(self):
+        build_vars = self.buildVars()
+        CodImpuesto1 = build_vars['CodImpuesto1']
+        CodImpuesto2 = build_vars['CodImpuesto2']
+        CodImpuesto3 = build_vars['CodImpuesto3']
+        return [
+            '%s' % build_vars['NumFac'],
+            '%s' % build_vars['FecFac'],
+            '%s' % build_vars['HoraFac'],
+            form.Amount(build_vars['ValorBruto']).format('%.02f'),
+            CodImpuesto1,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto1, 0.0)).format('%.02f'),
+            CodImpuesto2,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto2, 0.0)).format('%.02f'),
+            CodImpuesto3,
+            form.Amount(build_vars['ValorImpuestoPara'].get(CodImpuesto3, 0.0)).format('%.02f'),
+            form.Amount(build_vars['ValorTotalPagar']).format('%.02f'),
+            '%s' % build_vars['NitOFE'],
+            '%s' % build_vars['NumAdq'],
+            '%s' % build_vars['Software-PIN'],
+            '%d' % build_vars['TipoAmb'],
+        ]
+
 class DianXMLExtensionSoftwareProvider(FachoXMLExtension):
     # RESOLUCION 0004: pagina 108
 
@@ -167,7 +245,7 @@ class DianXMLExtensionSoftwareProvider(FachoXMLExtension):
         self.id_software = id_software
 
     def build(self, fexml):
-        software_provider = fexml.fragment('/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:SoftwareProvider')
+        software_provider = fexml.fragment('./ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:SoftwareProvider')
         provider_id_attrs = SCHEME_AGENCY_ATTRS.copy()
         provider_id_attrs.update({'schemeID': self.dv})
         #DIAN 1.7.-2020: FAB23
@@ -178,7 +256,6 @@ class DianXMLExtensionSoftwareProvider(FachoXMLExtension):
                                       **SCHEME_AGENCY_ATTRS)
 
 
-
 class DianXMLExtensionSoftwareSecurityCode(FachoXMLExtension):
     # RESOLUCION 0001: pagina 535
 
@@ -188,7 +265,7 @@ class DianXMLExtensionSoftwareSecurityCode(FachoXMLExtension):
         self.invoice_ident = invoice_ident
 
     def build(self, fexml):
-        dian_path = '/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:SoftwareSecurityCode'
+        dian_path = './ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:SoftwareSecurityCode'
         code = str(self.id_software) + str(self.pin) + str(self.invoice_ident)
         m = hashlib.sha384()
         m.update(code.encode('utf-8'))
@@ -216,7 +293,7 @@ class DianXMLExtensionSigner:
 
         fachoxml = FachoXML(xml,nsmap=NAMESPACES)
         #DIAN 1.7.-2020: FAB01
-        extcontent = fachoxml.builder.xpath(fachoxml.root, '/fe:Invoice/ext:UBLExtensions/ext:UBLExtension[2]/ext:ExtensionContent')
+        extcontent = fachoxml.builder.xpath(fachoxml.root, './ext:UBLExtensions/ext:UBLExtension[2]/ext:ExtensionContent')
         fachoxml.append_element(extcontent, signature)
 
         return fachoxml.tostring(xml_declaration=True, encoding='UTF-8')
@@ -280,22 +357,33 @@ class DianXMLExtensionSigner:
 
     def build(self, fachoxml):
         signature = self.sign_xml_element(fachoxml.root)
-        extcontent = fachoxml.builder.xpath(fachoxml.root, '/fe:Invoice/ext:UBLExtensions/ext:UBLExtension[2]/ext:ExtensionContent')
+        extcontent = fachoxml.builder.xpath(fachoxml.root, './ext:UBLExtensions/ext:UBLExtension[2]/ext:ExtensionContent')
         fachoxml.append_element(extcontent, signature)
 
 class DianXMLExtensionAuthorizationProvider(FachoXMLExtension):
     # RESOLUCION 0004: pagina 176
 
     def build(self, fexml):
-        dian_path = '/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:AuthorizationProvider/sts:AuthorizationProviderID'
-        fexml.set_element(dian_path, '800197268')
-
         attrs = {'schemeID': '4', 'schemeName': '31'}
         attrs.update(SCHEME_AGENCY_ATTRS)
-        fexml.set_attributes(dian_path, **attrs)
+        
+        authorization_provider = fexml.fragment('./ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:AuthorizationProvider')
+        authorization_provider.set_element('./sts:AuthorizationProviderID',
+                                           '800197268',
+                                           **attrs)
 
 
 
+class DianXMLExtensionInvoiceSource(FachoXMLExtension):
+    # CAB13
+    def build(self, fexml):
+        dian_path = '/fe:CreditNote/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:InvoiceSource/cbc:IdentificationCode'
+        fexml.set_element(dian_path, 'CO',
+                          listAgencyID="6",
+                          listAgencyName="United Nations Economic Commission for Europe",
+                          listSchemeURI="urn:oasis:names:specification:ubl:codelist:gc:CountryIdentificationCode-2.1")
+
+
 class DianXMLExtensionInvoiceAuthorization(FachoXMLExtension):
     # RESOLUCION 0004: pagina 106
 
@@ -310,17 +398,7 @@ class DianXMLExtensionInvoiceAuthorization(FachoXMLExtension):
         self.to = to
 
     def build(self, fexml):
-        fexml.set_element('/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:InvoiceSource/cbc:IdentificationCode',
-                          'CO',
-                          #DIAN 1.7.-2020: FAB15
-                          listAgencyID="6",
-                          #DIAN 1.7.-2020: FAB16
-                          listAgencyName="United Nations Economic Commission for Europe",
-                          #DIAN 1.7.-2020: FAB17
-                          listSchemeURI="urn:oasis:names:specification:ubl:codelist:gc:CountryIdentificationCode-2.1"
-                          )
-
-        invoice_control = fexml.fragment('/fe:Invoice/ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:InvoiceControl')
+        invoice_control = fexml.fragment('./ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:InvoiceControl')
         invoice_control.set_element('/sts:InvoiceControl/sts:InvoiceAuthorization', self.authorization)
         invoice_control.set_element('/sts:InvoiceControl/sts:AuthorizationPeriod/cbc:StartDate',
                                     self.period_startdate.strftime('%Y-%m-%d'))
@@ -333,6 +411,16 @@ class DianXMLExtensionInvoiceAuthorization(FachoXMLExtension):
         invoice_control.set_element('/sts:InvoiceControl/sts:AuthorizedInvoices/sts:To',
                                     self.to)
 
+        fexml.set_element('./ext:UBLExtensions/ext:UBLExtension/ext:ExtensionContent/sts:DianExtensions/sts:InvoiceSource/cbc:IdentificationCode',
+                          'CO',
+                          #DIAN 1.7.-2020: FAB15
+                          listAgencyID="6",
+                          #DIAN 1.7.-2020: FAB16
+                          listAgencyName="United Nations Economic Commission for Europe",
+                          #DIAN 1.7.-2020: FAB17
+                          listSchemeURI="urn:oasis:names:specification:ubl:codelist:gc:CountryIdentificationCode-2.1"
+                          )
+
 
 
 class DianZIP:
